/*
 * Netris -- A free networked version of T*tris
 * Copyright (C) 1994,1995,1996  Mark H. Weaver <mhw@netris.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "netris.h"

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <errno.h>
#include <setjmp.h>

#include "inet.h"

#define HEADER_SIZE sizeof(netint2[2])
#define HEADER_SIZE3 sizeof(netint4[3])

static MyEventType NetGenFunc(EventGenRec *gen, MyEvent *event);
static EventGenRec netGen = {
	NULL, 0, FT_read, -1, NetGenFunc, EM_net, 0, "\0", 0, HEADER_SIZE3
};


static sigjmp_buf close_env;

static void CatchInt(int sig)
{
	siglongjmp(close_env, 1);
}

int InitiateConnection(char *hostStr, short port)
{ //connect to host
	struct sockaddr_in addr;
	struct hostent *host;

	if (sigsetjmp(close_env, 1))
		exit(0);
	signal(SIGINT, CatchInt);  //handle exit (^C)
	AtExit(CloseNet);
	host = gethostbyname(hostStr);
	if (!host)
		die("gethostbyname");
	assert(host->h_addrtype == AF_INET);
 again:
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = host->h_addrtype;
	memcpy(&addr.sin_addr, host->h_addr, host->h_length);
	addr.sin_port = htons(port);
	if ((netGen.fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		die("socket");
	if (connect(netGen.fd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		if (errno != ECONNREFUSED)
			die("connect");
		close(netGen.fd);
		sleep(1);
		goto again;
	}
	AddEventGen(&netGen);
	return 0;
}

void HandShake(void)
{ //talk to your host
	MyEvent event;

	{
		netint4 versiondata[2];
		versiondata[0] = hton4(MAJOR_VERSION);
		versiondata[1] = hton4(PROTOCOL_VERSION);
		SendPacket(0, NP_hello, sizeof(versiondata), versiondata);
	}

	do {
		if (WaitMyEvent(&event, EM_net) == E_net)
			switch (event.u.net.type) {
			case NP_hello:
			{
				me = event.u.net.uid;
				memcpy(&Players[me], &Players[0], sizeof(player_t));
				fprintf(stderr, "Accepted (%s) as #%d (%s)\n",
					event.u.net.data, me, Players[me].name);
				SendPacket(0, NP_newPlayer,
					sizeof(player_t) - sizeof(Players[me].host),
					&Players[me]);
				break;
			}
			case NP_team:
			{ //receive your teamnumber
				memcpy(&Players[event.u.net.uid].team, event.u.net.data,
					event.u.net.size);
				break;
			} //NP_team
			case NP_gamedata:
			{
				static struct {
					int playerflags;
					int gravity;     //1
					int started;     //2
					int continuous;  //3
					long seed;       //4
					float shapes[7];
					int initspeed;   //5
				} data;

				memcpy(&data, event.u.net.data, event.u.net.size);
				memcpy(&Players[me].flags, &data, sizeof(data.playerflags));
				memcpy(&Players[me].flags, &data, sizeof(data.playerflags));
				memcpy(&Game.gravity, &data.gravity,
					sizeof(data) - sizeof(data.playerflags));
				break;
			} //NP_gamedata
			case NP_error:
			{
				fprintf(stderr, "Rejected by server: %s\n", event.u.net.data);
				exit(1);
			} //NP_error
			default:
				break;
			}
		else
			fatal("Hm, the party apparantly ended prematurely.");
	} while (event.u.net.type != NP_gamedata);
}


static MyEventType NetGenFunc(EventGenRec *gen, MyEvent *event)
{ //receive
	int result;
	short uid, type, size;
	netint4 *data = (netint4*)&(gen->buf);

	result = MyRead(gen->fd, gen->buf + gen->bufSize,
		gen->bufGoal - gen->bufSize);
	if (result < 0) {
		close(gen->fd);
		gen->fd = -1;
		return E_lostConn;
	}
	gen->bufSize += result;
	if (gen->bufSize < gen->bufGoal)
		return E_none;
	// *ugly* memcpy(data, gen->buf, sizeof(data));
	uid = ntoh4(data[0]);
	type = ntoh4(data[1]);
	size = ntoh4(data[2]);
	gen->bufGoal = size;
	if (gen->bufSize < gen->bufGoal)
		return E_none;
	gen->bufSize = 0;
	gen->bufGoal = HEADER_SIZE3;
	event->u.net.sender = gen->player;
	event->u.net.uid = uid;
	event->u.net.type = type;
	event->u.net.size = size - HEADER_SIZE3;
	event->u.net.data = gen->buf + HEADER_SIZE3;
	if (type == NP_endConn) return E_lostConn;
	return E_net;
}

void SendPacket(short uid, NetPacketType type, int size, void *data)
{ //send shit to server
	netint4 header[3];

	header[0] = hton4(uid);
	header[1] = hton4(type);
	header[2] = hton4(size + HEADER_SIZE3);
	if (MyWrite(netGen.fd, header, HEADER_SIZE3) != HEADER_SIZE3)
		die("write (header)");
	if (size > 0 && data && MyWrite(netGen.fd, data, size) != size)
		die("write");
}

void CloseNet(void)
{ //kick some connection's ass!
	MyEvent event;

	if (netGen.fd >= 0) {
		SendPacket(0, NP_endConn, 0, NULL);
		close(netGen.fd);
		netGen.fd = -1;
	}
	if (netGen.next)
		RemoveEventGen(&netGen);
}

